Require Import ssreflect ssrbool eqtype seq choice ssrnat.
Require Import libs.edone libs.bcase libs.base libs.fset tableaux.

Set Implicit Arguments.
Variable (T : choiceType).


(** examples listed in thesis **)
Section Thesis.
  
Example ex_thesis_a (A B : {fset T}) (p : T->bool) :
  let C := [fset X in powerset B | X `<=` [fset y in X | p y]] in
  A `<=` B -> [fset x in A | p x] \in C.
Proof. fset_dec. Qed.

Example ex_thesis_b1 (A B C: {fset T}) : A`<=`C -> B`<=`C -> ((C`\`A) `|` (C`\`B)) == (C`\`(A`&`B)).
Proof. fset_dec. Qed.

Example ex_thesis_b2 (A B C: {fset T}) : A`<=`C -> B`<=`C -> ((C`\`A) `&` (C`\`B)) == (C`\`(A`|`B)).
Proof. fset_dec. Qed.

(*
takes longer as stated in the thesis because the application order of the cut rules
has changed. As membership cuts are neeeded more often than equality cuts,
they were moved to the front. The currently measured time is 45.43 sec.
*)
Example ex_thesis_c (A B : {fset T}):
  (powerset (A`|`B) `<=` powerset A `|` powerset B) -> (A `<=` B) || (B `<=` A).
Proof. fset_dec. Qed.

Example ex_thesis_d (x:T) A B : A `\` B `<=` fset0 -> x \in A -> x \in B.
Proof. fset_dec. Qed.

Example ex_thesis_e (A B : {fset T}) : powerset A `<=` powerset B -> A `<=` B.
Proof. fset_dec. Qed.

(* Same problem as above. Currently measured time: 15.25 sec. *)
Example ex_thesis_f (x:T) A : (powerset A) `\` [fset A] `<=` fset0 -> A == fset0 :> {fset T}.
Proof. fset_dec. Qed.

Variables (p q r : pred {fset T}) (F : {fset T}).
Definition U := powerset F.
Definition S0 := [fset D in U | p D && q D].
Definition A := [fset D in S0 | r D].

Example ex_thesis_g s C : s \in C -> C \in A -> s \in F.
Proof. fset_decu (A,S0,U). Qed.

End Thesis.


(*
Example for an infinite saturation process.
Corresponds to the branch in Section 4.4.2
Diverges even without cuts.
 *)
Example nonterm (x:T) (A B C : {fset T}):
  let F := [fset x in A | ~~ (B `<=` x |` C)] in x \in F -> B `<=` F -> False.
Proof. (* fset_nocut. *) (* diverges *) Abort.


(** Examples solved in negligible time **)

Example p_00 (A B C : {fset T}): A`<=`B -> A`<=`C -> A`<=`(B`&`C).
fset_dec. Qed.

Example p_01 (A B C : {fset T}) : A`&`B`&`C`<=`C`&`A.
fset_dec. Qed.

Example p_02 (x:T) (A:{fset T}) : [fset x]`|`A `<=` A -> x \in A.
fset_dec. Qed.

Example p_03 (A B : {fset T}) : A `|` B `<=` A -> B `<=` A.
fset_dec. Qed.

(* note that Coq-equalities can be processed as well. Not only == *)
Example p_04 (x y z : T) : [fset x in [fset y] | x \in [fset z]] != fset0 -> y=z.
fset_dec. Qed.

Example p_05 (A B C : {fset T}) : (A`\`B)`\`C == A `\` (B `|` C).
fset_dec. Qed.

Example p_06 (A B C : {fset T}) : A `\` (B`\`C) == (A`\`B) `|` (A`&`C).
fset_dec. Qed.

Example p_07 (A B : {fset T}) : A != fset0 -> A`<=`B -> A`&`B != fset0.
fset_dec. Qed.

Example p_08 (A B C : {fset T}) : A \in [fset X in (powerset B) | C `<=` X] -> C `<=` B.
fset_dec. Qed.

Example p_09 (A B C : {fset T}) : (A`<=`B) && (B`<=`C) -> C`<=`A -> (A==B) && (B==C) && (A==C).
fset_dec. Qed.

Example p_10 (x:T) (A:{fset T}) : [fset x]`&`A  == [fset x] -> x \in A.
fset_dec. Qed.

Example p_11 (A B : {fset T}): powerset A `|` powerset B `<=` powerset (A`|`B).
fset_dec. Qed.

Example p_12 (A B : {fset T}): powerset (A`&`B) == powerset A `&` powerset B.
fset_dec. Qed.


(** examples using unfold **)

Section Unfold.
  
Variable A' B' : {fset T}.
Definition F := powerset A'.
Definition G := F `|` powerset B'.

Example u_00 (X : {fset T}) : X `<=` A' -> X \in G. fset_decu (G, F). Qed.


Variable (K L : {fset T}) (p : {fset T} -> bool).
Definition M := powerset L.
Definition N := sep M p.

Example u_01 (s:T) : K \in N -> s \in L -> K `|` [fset s] \in M. fset_decu (N, M). Qed.

End Unfold.


(** larger examples **)

Definition gen_set (n:nat) : {fset T}. Admitted.

Function trans_chain n :=
  match n with
    | 0 => ((gen_set 1) `<=` (gen_set 0))
    | S m =>
      let H := trans_chain m in
      ((gen_set (S n) `<=` (gen_set n)) && H)
  end.

(* single branch *)

(* 3.93 sec *)
Example l_00 (x:T) : x \in (gen_set 20) -> trans_chain 20 -> x \in (gen_set 0).
fset_decu trans_chain. Qed.

(* 7.82 sec *)
Example l_01 : trans_chain 20 -> gen_set 0 `<=` gen_set 21 -> gen_set 5 = gen_set 15.
fset_decu trans_chain.


