Require Import ssreflect ssrbool eqtype ssrnat seq.
Require Import taut.

Opaque implb andb orb negb.


(** example generators **)

Definition ass : nat -> bool. Admitted.

(** ** tautologies ** **)

Function XMple_conj n :=
  match n with
  | 0 => (ass 0) || ~~ (ass 0)
  | n.+1 => ((ass n.+1) || ~~(ass n.+1)) && XMple_conj n
  end.

Function XMple_disj n :=
  match n with
    | 0 => false
    | n.+1 => ((ass n) || ~~(ass n)) || XMple_disj n
  end.

Function and_count b n :=
  match n with
    | 0 => b || ass 0
    | S n' => and_count b n' && (b || ass n)
  end.

(*
Function or_id n :=
  match n with
    | 0 => true
    | S n => (and_var true n || (ass 0 == ass 0)) && or_id n
  end.
 *)

Definition impl_chain n :=
  let m := n in
  let f := (fix ic m :=
              match m with
              | 0 => ass n ==> ass n.+1
              | m.+1 =>
                let k := n-m.+1 in
                (ass k ==> ass k.+1) && ic m
              end) in
  ass 0 && f m ==> ass n.+1.

Definition impl_chain2 n :=
  let m := n in
  let f := (fix ic m :=
              match m with
                | 0 => (ass n ==> ass n.+1) && (~~ ass n ==> ass n.+1)
                | m.+1 =>
                  let k := n-m.+1 in
                  (ass k ==> ass k.+1) && (~~ ass k ==> ass k.+1) && ic m
              end) in
  f m ==> ass n.+1. 


(** ** conditional tautologies (depending on the input) ** **)

(* tautology for b = true *)
Function XMple_conj_hide b n :=
  match n with
  | 0 => b
  | n.+1 => ((ass n) || ~~(ass n)) && XMple_conj_hide b n
  end.

(* tautology for b=true *)
Function or_var b n :=
  match n with
    | 0 => b
    | n.+1 => (ass n || or_var b n)
  end.


(** ** building blocks ** **)

Function and_k_from k m :=
  match k with
    | 0 => ass m
    | S k' => ass (m + k) && and_k_from k' m
  end.

Function or_nt n :=
  match n with
    | 0 => ass 0 == ass 0
    | S n => and_k_from 25 (n * 25) || or_nt n
  end.


(** ** non-tautology ** **)

Function and_var b n :=
  match n with
    | 0 => b
    | n.+1 => (ass n && and_var b n)
  end.

Function onevar_conj n b :=
  match n with
    | 0 => b
    | n.+1 => b && onevar_conj n b
  end.

Function bal_and_ced n m :=
  match n with
    | 0 => ass m
    | n.+1 => (bal_and_ced n m.+1) && (bal_and_ced n (m + 2^n))
  end.

Function bal_or_ced n m:=
  match n with
    | 0 => ass m
    | n.+1 => (bal_or_ced n (m.+1)) || (bal_or_ced n (m + 2^n))
  end.

Definition impl_chainF n :=
  let m := n in
  let f := (fix ic m := match m with
                          | 0 => ass n ==> ass n.+1
                          | m.+1 =>
                            let k := n-m.+1 in
                            (ass k ==> ass k.+1) && ic m
                        end) in
  ass 0 && f m ==> ass n.+2.


(** ** bool-to-Prop translation ** **)

Ltac b2P b :=
  match b with
  | true => True
  | false => False
  | ~~ ?b1 =>  let P1 := b2P b1 in
               constr:(~ P1)
  | ?b1 && ?b2 => let P1 := b2P b1 in
                  let P2 := b2P b2 in
                  constr:(P1 /\ P2)
  | ?b1 || ?b2 => let P1 := b2P b1 in
                  let P2 := b2P b2 in
                  constr:(P1 \/ P2)
  | ?b1 == ?b2 => let P1 := b2P b1 in
                  let P2 := b2P b2 in
                  constr:(P1 = P2)
  | ?b1 ==> ?b2 => let P1 := b2P b1 in
                   let P2 := b2P b2 in
                   constr:(P1 -> P2)
  | _ => constr:(b)
  end.

Ltac gtp :=
  let V := fresh "H" in
  match goal with
  | |- is_true ?C => let P := b2P C in have V : ~ ~ P
  | |- ?C = true => let P := b2P C in have V : ~ ~ P
  end.


(** Examples listed in thesis **)

Example XM : XMple_conj 9.
unfold XMple_conj.
gtp. firstorder. clear H. (* 2.97 s *)
gtp. tauto. (* > 60 s *)
shandec. (* 2.25 s *)
Qed.

Example ICF : impl_chainF 40.
unfold impl_chainF.
gtp. try (firstorder; fail). (* 6.97 s *)
try tauto. (* 3.85 s *) admit.
try shandec. (* 2. 29 s *)
Abort.

Example IC2 : impl_chain2 40.
unfold impl_chain2; cbv.
gtp. firstorder. clear H. (* 2.37 s *)
gtp. tauto. (* 8.60 s *)
shandec. (* 2.81 s *)
Qed.

Example IC : impl_chain 60.
unfold impl_chain; cbv.
gtp. firstorder. clear H. (* 4.00 s *)
gtp. tauto. (* 4.60 s *)
shandec. (* 2.06 s *)
Qed.

(** proof term comparison between tauto, firstorder and taut **)

Example pt_0 a b :  ~ ~ ((a \/ ~ a) /\ (b \/ ~ b)).
firstorder. Show Proof. Qed.

Example pt_1  a b :  ~ ~ ((a \/ ~ a) /\ (b \/ ~ b)).
tauto. Show Proof. Qed.

Example pt_2 a b: ~~ ~~ ((a || ~~ a) && (b || ~~ b)).
shandec. Show Proof. Qed.


(** further large examples **)

(* binary disjunction tree with pairwise distinct vars *)
Example D1 : bal_or_ced 6 1.
unfold bal_or_ced.
gtp. try (firstorder; fail). (* 6.97 s *)
try tauto. admit. (* 1.32 s *)
try shandec. (* 1.22s *)
Abort.

(* linear disjunction of 30 (a_i || ~~ a_i) statements*)
Example D2a : XMple_disj 30.
unfold XMple_disj.
gtp. firstorder. clear H. (* 0.78 s *)
gtp. tauto. (* 1.12 s *)
shandec. (* ~ 0 s *)
Qed.

(* linear conjunction of 10 (a_i || ~~ a_i) statements with true at the end*)
Example D2b : XMple_conj_hide true 10.
unfold XMple_conj_hide.
gtp. firstorder. clear H. (* 4.65 s *)
gtp. (*tauto.*) admit. (* >60 s *)
shandec. (* 6.00 s *)
Qed.

(* linear disjunction of 70 vars with true at the end *)
Example D3 : or_var true 70.
unfold or_var.
gtp. firstorder. clear H. (* 5.84 s *)
gtp. tauto. (* 23.94 s *)
shandec. (* 2.22 s *)
Qed.

(* linear conjunction of 50 vars with a true at the end *)
Example D4 : and_var true 50.
unfold and_var.
gtp. try (firstorder; fail). (* ~ 0 s *)
try tauto. (* ~ 0 s *)
try shandec. (* ~ 0 s *)
Abort.

(* disjunction of 10 blocks of conjunctions á 25 distinct vars and a trivial tautology at the end *)
Example D5 : or_nt 4.
unfold or_nt, and_k_from.
gtp. firstorder. clear H. (* 2.75 s *)
gtp. tauto. (* 1.35 s *)
(*shandec.*) (* > 60 s *)
Abort.

(* conjunction of 8 disjunctions of (a || ~~a) with pairwise distinct vars *)
Example D6 : and_count (ass 10 || ~~ ass 10) 8.
unfold and_count.
gtp. firstorder. clear H. (* 3.87 s *)
(*gtp. tauto.*) (* 51.75 s *)
shandec. (* ~ 0 s *)
Qed.

(* negation of a linear conjunction of 100 distinct vars with a false at the end*)
Example D7: ~~ and_var false 100.
unfold and_var.
gtp. firstorder. clear H. (* 7.53 s *)
gtp. tauto. (* 4.94 s *)
shandec. (* 1.44 s *)
Qed.

Example D11 : impl_chain 50.
unfold impl_chain; cbv.
gtp. firstorder. clear H. (* 2.18 s *)
gtp. tauto. (* 2.63 s *)
shandec. (* 1.50 s *)
Qed.

Example D12 : impl_chainF 50.
unfold impl_chainF; cbv.
gtp. try (firstorder; fail). (* 5.35 s *)
try tauto. admit. (* 2.71 s *)
try shandec. (* 1.40 s *)
Abort.

(*
* Results:
*
* compared with tauto/firstorder taut is ...
*   - faster when constants can be propagated
*   - faster when many logical operations are involved
*   - slower when many variables are involved
*)